/*
 * BDM flash utility for the Coldfire Flash Module on MCF5282
 *
 * Copyright (C) 2005 Zachary P. Landau <kapheine@hypa.net>
 *
 * Heavily based on uClinux-2.4.x/drivers/mtd/devices/cfm.c by Mathias Kster
 * m5282sim.h taken from uClinux-2.4.x/include/asm-m68knommu 
 *
 * Date: 2005-02-09
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "cfmflash_lo.h"

/* #define DEBUG printf */
#define DEBUG(...)

static char *program_name = "cfmflash";
static char *version = "0.1";
static char *copyright = "Copyright (C) 2005 Zachary P. Landau";

static char bdm_device[256] = "/dev/bdmcf0";
static int verbose;
static int no_verify = 0;

void show_usage()
{
    printf("Usage: [-d, --device=BDMDEVICE] [-h, --help] [-v, --verbose] [-V, --version] <command> [options...]\n");
    printf("  info\t\t\t\t\tdisplay driver information\n");
    printf("  erase\t\t\t\t\terase entire flash\n");
    printf("  erase_region <region> <length]\terase any sectors within region-length\n");
    printf("  verify\t\t\t\tverify flash is erased\n");
    printf("  verify_region <region> <length>\tverify any sectors within region-length are erased\n");
    printf("  write <filename> <region>\t\twrite file into flash starting at region\n");
    printf("  read <filename> <region> <length>\tread length bytes from region into filename\n");
    exit(EXIT_FAILURE);
}

void show_version()
{
    printf("%s %s\n%s\n", program_name, version, copyright);
}

void check_command_args(int optch, int number)
{
    if (optch-1 < number)
        show_usage();
}

int command_info(void)
{
    flash_init(bdm_device);
    flash_info();
    flash_shutdown();
    
    return 0;
}

int command_erase(void)
{
    flash_init(bdm_device);
    if (verbose)
        printf("Erasing...");
    flash_erase();
    if (verbose)
        printf("done\n");
    flash_shutdown();

    return 0;
}

int command_verify(void)
{
    flash_init(bdm_device);
    fprintf(stderr, "Verifying...");
    if (flash_verify())
        fprintf(stderr, "not ");
    fprintf(stderr, "erased\n");
    flash_shutdown();

    return 0;
}

int command_erase_region(unsigned long region, unsigned long length)
{
    flash_init(bdm_device);
    flash_erase_region(region, length, verbose);
    flash_shutdown();

    return 0;
}

int command_verify_region(unsigned long region, unsigned long length)
{
    flash_init(bdm_device);
    flash_verify_region(region, length, 1);
    flash_shutdown();

    return 0;
}

int command_write(const char *filename, unsigned long region)
{
    unsigned char buffer[FLASH_BLOCKSIZE];
    FILE *fp;
    unsigned long bytes;

    fp = fopen(filename, "r");
    if (!fp) {
        fprintf(stderr, "could not open %s\n", filename);
        return -1;
    }

    flash_init(bdm_device);

    while ((bytes = fread(&buffer, 1, sizeof(buffer), fp)) != 0) {
        /* erase first, if we want to */
        if (!no_verify && flash_verify_region(region, FLASH_BLOCKSIZE-1, verbose))
            fprintf(stderr, "WARNING: region %#lx not erased\n", region);

        if (verbose)
            fprintf(stderr, "Writing %#lx-%#lx...", region, region+FLASH_BLOCKSIZE-1);
        flash_write(region, bytes, buffer);
        if (verbose)
            fprintf(stderr, "done\n");
        region += bytes;
    }

    flash_shutdown();

    fclose(fp);

    return 0;
}

int command_read(const char *filename, unsigned long region, unsigned long length)
{
    unsigned char buffer[FLASH_BLOCKSIZE];
    FILE *fp;
    unsigned long pos;

    fp = fopen(filename, "w");
    if (!fp) {
        fprintf(stderr, "could not open %s\n", filename);
        return -1;
    }

    flash_init(bdm_device);

    for (pos = region; pos <= region + length; pos += FLASH_BLOCKSIZE) {
        if (verbose)
            fprintf(stderr, "Reading %#lx-%#lx...", pos, region+FLASH_BLOCKSIZE-1);
        flash_read(pos, sizeof(buffer), buffer);
        if (verbose)
            fprintf(stderr, "done\n");
        fwrite(buffer, 1, sizeof(buffer), fp);
    }

    flash_shutdown();
    fclose(fp);

    return 0;
}

static const char *optstring = "ed:hvV";
static struct option long_options[] = {
    { "device",     1,  NULL,   'd' },
    { "erase",      0,  NULL,   'e' },
    { "help",       0,  NULL,   'h' },
    { "no-verify",  0,  NULL,   'n' },
    { "verbose",    0,  NULL,   'v' },
    { "version",    0,  NULL,   'V' },
    { NULL }
};

int main(int argc, char **argv)
{
    int optch;
    char *command, *filename;
    unsigned long region;
    unsigned long length;

    int erase = 0;

    while ((optch = getopt_long(argc, argv, optstring, long_options, NULL)) != -1) {
        switch (optch) {
            case 'd':
                strncpy(bdm_device, optarg, sizeof(bdm_device));
                break;
            case 'e':
                erase = 1;
                break;
            case 'h':
                show_usage();
                break;
            case 'n':
                no_verify = 1;
                break;
            case 'v':
                verbose = 1;
                break;
            case 'V':
                show_version();
                return EXIT_SUCCESS;
                break;
        }
    }

    /* optch - number of arguments left argv[optind] is the first */

    optch = argc - optind;
    if (optch < 1)
        show_usage();

    command = argv[optind];
  
    if (strcasecmp(command, "erase_region") == 0) {
        check_command_args(optch, 2);
        region = strtoul(argv[optind+1], NULL, 0);
        length = strtoul(argv[optind+2], NULL, 0);

        DEBUG("erase %#lx %#lx\n", region, length);
        command_erase_region(region, length);
    } else if (strcasecmp(command, "erase") == 0) {
        DEBUG("erase\n");
        command_erase();
    } else if (strcasecmp(command, "verify") == 0) {
        DEBUG("verify\n");
        command_verify();
    } else if (strcasecmp(command, "verify_region") == 0) {
        check_command_args(optch, 2);
        region = strtoul(argv[optind+1], NULL, 0);
        length = strtoul(argv[optind+2], NULL, 0);

        DEBUG("verify_region %#lx %#lx\n", region, length);
        command_verify_region(region, length);
    } else if (strcasecmp(command, "write") == 0) {
        check_command_args(optch, 2);
        filename = argv[optind+1];
        region = strtoul(argv[optind+2], NULL, 0);

        DEBUG("write %s %#lx\n", filename, region);
        command_write(filename, region);
    } else if (strcasecmp(command, "read") == 0) {
        check_command_args(optch, 3);
        filename = argv[optind+1];
        region = strtoul(argv[optind+2], NULL, 0);
        length = strtoul(argv[optind+3], NULL, 0);

        DEBUG("read %s %#lx %#lx\n", filename, region, length);
        command_read(filename, region, length);
    } else if (strcasecmp(command, "info") == 0) {
        DEBUG("info\n");

        command_info();
    } else {
        show_usage();
    }
 
    return EXIT_SUCCESS;

}
