/*
 * BDM flash utility for the Coldfire Flash Module on MCF5282
 *
 * Copyright (C) 2005 Zachary P. Landau <kapheine@hypa.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "cfmflash_lo.h"

static void reset_state(void)
{
      unsigned char ustat;

      bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
      if (ustat & CFMUSTAT_PVIOL)
          bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, ustat | CFMUSTAT_PVIOL);
      //bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &byte);
      if (ustat & CFMUSTAT_ACCERR)
          bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, ustat | CFMUSTAT_ACCERR);
      //bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &byte);
      if (ustat & CFMUSTAT_BLANK)
          bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, ustat | CFMUSTAT_BLANK);
}

static void cfm_flash_init(void)
{
    unsigned long longword;

    reset_state();
    bdmWriteWord(MCF_IPSBAR+MCF5282_CFM_MCR, 0);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_CLKD, (1<<6) | 20);
    bdmWriteLongWord(MCF_IPSBAR+MCF5282_CFM_PROT, 0);
    bdmWriteLongWord(MCF_IPSBAR+MCF5282_CFM_SACC, 0);
    bdmWriteLongWord(MCF_IPSBAR+MCF5282_CFM_DACC, 0);

    bdmReadLongWord(MCF_IPSBAR+MCF5282_CFM_SEC, &longword);
    if (longword & 0x80000000)
        fprintf(stderr, "Warning: backdoor mode enabled. (%#lx)\n", longword);
    if (longword & 0x40000000)
        fprintf(stderr, "Warning: security mode enabled. (%#lx)\n", longword);
}

int flash_init(const char *bdm_device)
{
    if (bdmOpen(bdm_device) < 0) {
        fprintf(stderr, "Could not open device %s\n", bdm_device);
        return -1;
    }

    cfm_flash_init();

    return 0;
}

int flash_shutdown(void)
{
    return bdmClose();
}

int flash_info(void)
{
    unsigned int ver;
    unsigned int cpu;
    unsigned int status;

    if (bdmGetDrvVersion(&ver) < 0) {
        fprintf(stderr, "GetDrvVersion\n");
        return -1;
    }
    printf("Driver Version: %i.%i\n", ver >> 8, ver & 0xff);

    if (bdmGetProcessor(&cpu) < 0) {
        fprintf(stderr, "bdmGetProcessor\n");
        flash_shutdown();
        return -1;
    }

    switch (cpu) {
        case BDM_COLDFIRE:
            printf("Processor: ColdFire\n");
            break;
        default:
            printf("Unsupported processor\n");
            flash_shutdown();    
            return -1;
            break;
    }

    status = bdmStatus();
    printf("Target status: 0x%x -- %s, %s, %s, %s, %s.\n", status,
           status & BDM_TARGETRESET   ? "RESET" : "NOT RESET",
           status & BDM_TARGETHALT    ? "HALTED" : "NOT HALTED",     
           status & BDM_TARGETSTOPPED ? "STOPPED" : "NOT STOPPED",           
           status & BDM_TARGETPOWER   ? "POWER OFF" : "POWER ON",
           status & BDM_TARGETNC      ? "NOT CONNECTED" : "CONNECTED");     

    return 0;
}

int flash_erase(void) 
{
    unsigned char ustat;
    int status = 0;

    bdmWriteLongWord(CFM_BACKDOOR_OFFSET, 0);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_CMD, CFMCMD_MASERS);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, CFMUSTAT_CBEIF);

    do {
        bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
    } while (!(ustat & CFMUSTAT_CCIF));

    if (ustat & CFMUSTAT_ACCERR) {
        fprintf(stderr, "mass_erase: ACCERR\n");
        status = 1;
    }

    if (ustat & CFMUSTAT_PVIOL) {
        fprintf(stderr, "mass_erase: PVIOL\n");
        status = 1;
    }

    reset_state();

    return status;
}

int flash_verify(void) 
{
    unsigned char ustat;
    int status = 0;

    bdmWriteLongWord(CFM_BACKDOOR_OFFSET, 0);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_CMD, CFMCMD_RDARY1);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, CFMUSTAT_CBEIF);

    do {
        bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
    } while (!(ustat & CFMUSTAT_CCIF));

    status = (ustat & CFMUSTAT_BLANK) == 0;

    bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
    if (ustat & CFMUSTAT_ACCERR) {
        fprintf(stderr, "flash_verify: ACCERR\n");
        status = 1;
    }

    bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
    if (ustat & CFMUSTAT_PVIOL) {
        fprintf(stderr, "flash_verify: PVIOL\n");
        status = 1;
    }

    reset_state();

    return status;
}

static int erase_block(unsigned long region)
{
    unsigned char ustat;
    int status = 0;

    bdmWriteLongWord(CFM_BACKDOOR_OFFSET+region, 0);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_CMD, CFMCMD_PGERS);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, CFMUSTAT_CBEIF);

    do {
        bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
    } while (!(ustat & CFMUSTAT_CCIF));

    if (ustat & CFMUSTAT_ACCERR) {
        fprintf(stderr, "mass_erase: ACCERR\n");
        status = 1;
    }

    if (ustat & CFMUSTAT_PVIOL) {
        fprintf(stderr, "mass_erase: PVIOL\n");
        status = 1;
    }

    reset_state();

    return status;
}


int flash_erase_region(unsigned long region, unsigned long length, int verbose)
{
    unsigned long pos;
    if (region + length > FLASH_SIZE) return -1;
    if (region % FLASH_BLOCKSIZE) return -1;

    for (pos = region; pos <= region + length; pos += FLASH_BLOCKSIZE+1) {
        if (verbose)
            fprintf(stderr, "Erasing region %#lx...", pos);
        erase_block(pos);
        if (verbose)
            fprintf(stderr, "done\n");
    }

    return 0;
}

static int verify_block(unsigned long region)
{
    unsigned long status = 1;
    unsigned char ustat;

    bdmWriteLongWord(CFM_BACKDOOR_OFFSET+region, 0);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_CMD, CFMCMD_PGERSVER);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, CFMUSTAT_CBEIF);

    do {
        bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
    } while (!(ustat & CFMUSTAT_CCIF));

    status = (ustat & CFMUSTAT_BLANK) == 0;

    if (ustat & CFMUSTAT_ACCERR) {
        fprintf(stderr, "flash_verify_region: ACCERR\n");
        status = 0;
    }

    if (ustat & CFMUSTAT_PVIOL) {
        fprintf(stderr, "flash_verify_region: IVIOL\n");
        status = 0;
    }

    reset_state();

    return status;
}

int flash_verify_region(unsigned long region, unsigned long length, int verbose)
{
    unsigned long pos;
    int status;
    if (region + length > FLASH_SIZE) return -1;
    if (region % FLASH_BLOCKSIZE) return -1;

    for (pos = region; pos <= region + length; pos += FLASH_BLOCKSIZE) {
        if (verbose)
            fprintf(stderr, "Verifying %#lx-%#lx...", pos, pos+FLASH_BLOCKSIZE-1);
        status = verify_block(pos);
        if (verbose && status)
            fprintf(stderr, "not ");
        if (verbose)
            fprintf(stderr, "erased\n");
    }

    return status;
}

static inline int write_dword(unsigned long region, unsigned long x)
{
    unsigned long status = 1;
    unsigned char ustat;

    //printf("Pretending to write %#lx -> %#lx\n", x, to);
    //return 1;
    bdmWriteLongWord(CFM_BACKDOOR_OFFSET+region, x);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_CMD, CFMCMD_PGM);
    bdmWriteByte(MCF_IPSBAR+MCF5282_CFM_USTAT, CFMUSTAT_CBEIF);

    do {
        bdmReadByte(MCF_IPSBAR+MCF5282_CFM_USTAT, &ustat);
    } while (!(ustat & CFMUSTAT_CCIF));

    if (ustat & CFMUSTAT_ACCERR) {
        fprintf(stderr, "write_dword: ACCERR\n");
        status = 0;
    }

    if (ustat & CFMUSTAT_PVIOL) {
        fprintf(stderr, "write_dword: IVIOL\n");
        status = 0;
    }

    reset_state();

    return status;
}

int flash_write(unsigned long region, size_t length, unsigned char *buf)
{
    unsigned char tmp[4];
    int i, n;

    if (region & (BUSWIDTH - 1)) {
        unsigned long aligned = region & ~(BUSWIDTH - 1);
        int gap = region - aligned;

        i = n = 0;

        while (gap--) tmp[i++] = 0xFF;
        while (length && i < BUSWIDTH) tmp[i++] = buf[n++], length--;
        while (i < BUSWIDTH) tmp[i++] = 0xFF;

        if (!write_dword(aligned,htonl(*((unsigned long *)tmp)))) return -1;

        region += n;
        buf += n;
    }

    while (length >= BUSWIDTH) {
        if (!write_dword(region,htonl(*((unsigned long *)buf)))) return -1;

        region += BUSWIDTH;
        buf += BUSWIDTH;
        length -= BUSWIDTH;
    }

    if (length & (BUSWIDTH - 1)) {
        i = n = 0;

        while (length--) tmp[i++] = buf[n++];
        while (i < BUSWIDTH) tmp[i++] = 0xFF;

        if (!write_dword(region, htonl(*((unsigned long *)tmp)))) return -1;
    }

    return 0;
}

int flash_read(unsigned long region, size_t length, unsigned char *buf)
{
    unsigned long i;
    unsigned char val;

    for (i=0; i < length; i++) {
        bdmReadByte(CFM_BACKDOOR_OFFSET+region+i, &val);
        buf[i] = val;
    }
    return 0;
}
